#!/bin/sh
# the next line restarts using the interpreter \
exec wish "$0" "$@"

# copyright (C) 1997-98 Jean-Luc Fontaine (mailto:jfontain@mygale.org)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: moodss.tcl,v 1.92 1998/06/20 20:38:25 jfontain Exp $}

set applicationVersion 4.0

source getopt.tcl

proc printUsage {exitCode} {
    puts stderr "Usage: $::argv0 \[OPTION\]... \[MODULE\] \[MODULE\]..."
    puts stderr {  -f, --file       configuration file name (implies no modules)}
    puts stderr {  -h, --help       display this help and exit}
    puts stderr {  -p, --poll-time  poll time in seconds}
    puts stderr {  --version        output version information and exit}
    exit $exitCode
}

if {[catch\
    {set argv [parseCommandLineArguments {-f 1 --file 1 -h 0 --help 0 -p 1 --poll-time 1 --version 0} $argv arguments]}\
    message\
]} {
    puts stderr $message
    printUsage 1
}
foreach {short long} {-f --file -h --help -p --poll-time} {
    catch {set arguments($short) $arguments($long)}                                          ;# long version if present has priority
}

if {[info exists arguments(--version)]} {
    puts "moodss (a Modular Object Oriented Dynamic SpreadSheet) version $applicationVersion"
    exit
}
if {\
    ([info exists arguments(-f)]&&([llength $argv]>0))||(![info exists arguments(-f)]&&([llength $argv]==0))||\
    [info exists arguments(-h)]\
} {
    printUsage 1
}

proc loadModules {modules} {
    foreach module $modules {
        if {[info exists loaded($module)]} {
            puts stderr "$::argv0: module \"$module\" was specified more than once"
            exit 1
        }
        if {[catch {package require $module}]} {
            puts stderr "$::argv0: could not load package \"$module\""
            exit 1
        }
        set loaded($module) {}
        if {![info exists ${module}::data(indexColumns)]} {
            set ${module}::data(indexColumns) 0                                ;# use first column as single index column by default
        }
    }
}

proc setPollTimes {modules {commandLineTime {}}} {
    global pollTimes pollTime

    set default 0
    set minimum 0
    foreach module $modules {
        set times [set ${module}::data(pollTimes)]
        if {[llength $times]==0} {
            error "module $module poll times list is empty"
        }
        set time [lindex $times 0]
        if {$time>$default} {                                                                  ;# default value is the first in list
            set default $time                                                        ;# keep the greater default time of all modules
        }
        set times [lsort -integer $times]                                                                         ;# sort poll times
        set time [lindex $times 0]
        if {$time>$minimum} {
            set minimum $time                                                        ;# keep the greater minimum time of all modules
            set minimumModule $module
        }
        foreach time $times {                                        ;# poll times list is the combination of all modules poll times
            set data($time) {}
        }
    }
    set pollTimes [lsort -integer [array names data]]         ;# sort and restrict poll times above maximum module minimum poll time
    set pollTimes [lrange $pollTimes [lsearch -exact $pollTimes $minimum] end]
    set pollTime $default
    if {[string length $commandLineTime]>0} {                                           ;# eventually validate command line override
        if {$commandLineTime<$minimum} {
            puts stderr "$::argv0: minimum poll time is $minimum seconds for module $minimumModule"
            exit 1
        }
        set pollTime $commandLineTime
    }
}

proc modulesString {} {
    global modules

    for {set index 0} {$index<([llength $modules]-1)} {incr index} {
        append string "[lindex $modules $index], "
    }
    append string [lindex $modules $index]
    return $string
}


if {[catch {package require stooop 3.6}]} {
    source stooop.tcl                                                                     ;# in case stooop package is not installed
}
namespace import stooop::*
if {[catch {package require switched 1.4}]} {                                           ;# in case switched package is not installed
    source switched.tcl
}

if {[catch {package require scwoop 2.2}]} {
    source scwoop.tcl                                                                     ;# in case scwoop package is not installed
    source bindings.tcl
    source widgetip.tcl
    source arrowbut.tcl
    source spinent.tcl
    source panner.tcl
}

source record.tcl

if {[info exists arguments(-f)]} {                                                              ;# configuration file name specified
    set saveFile $arguments(-f)
    set initializer [new record -file $saveFile]
    record::read $initializer
    set modules [record::modules $initializer]
} else {
    set saveFile {}
    set modules $argv
}

lappend auto_path .                                                   ;# search in current directory sub-directories for development
loadModules $modules

if {[info exists arguments(-p)]} {                                                     ;# command line argument has highest priority
    setPollTimes $modules $arguments(-p)
} elseif {[info exists initializer]} {                                                                  ;# then stored configuration
    setPollTimes $modules [record::pollTime $initializer]
} else {                                                                                                       ;# use modules values
    setPollTimes $modules
}
set modulesString [modulesString]                                                      ;# store often used concatenated module names

if {[catch {package require Tktable 2.0}]} {                                             ;# in case tkTable package is not installed
    switch $tcl_platform(platform) {
        unix {
            load ./Tktable.so.2.00
        }
        windows {
            load ./tktable.dll
        }
    }
}

if {[catch {package require BLT}]} {                                                         ;# in case BLT package is not installed
    switch $tcl_platform(platform) {
        unix {                                                                      ;# try latest official version or unofficial 8.0
            if {[catch {load ./libBLT2.4.so} officialMessage]&&[catch {load ./libBLT8.0.so} message]} {
                puts $officialMessage
                puts $message
                exit 1
            }
        }
        windows {
            load ./blt80.dll                                                  ;# so far, only unofficial 8.0 is supported on Windows
        }
    }
}
set officialBLT [expr {$blt_version!=8.0}]                                               ;# remember which BLT version is being used

if {[catch {package require tkpiechart 5.2}]} {                                       ;# in case tkpiechart package is not installed
    source pielabel.tcl
    source boxlabel.tcl
    source canlabel.tcl
    source labarray.tcl
    source perilabel.tcl
    source slice.tcl
    source selector.tcl
    source objselec.tcl
    source pie.tcl
}

source utility.tcl
source config.tcl
source font.tcl
source scroll.tcl
source lifolbl.tcl
source keyslink.tcl
source dialog.tcl
source viewer.tcl
source blt2d.tcl
source tablesel.tcl
source datatab.tcl
source datagraf.tcl
source databar.tcl
source datapie.tcl
source htmllib.tcl                                                ;# Tcl HTML library from Sun, used for viewing HTML help documents
source htmldata.tcl
source html.tcl                                             ;# must be source after HTML library since some peocedures are redefined
source help.tcl
source drag.tcl
source drop.tcl
source canvhand.tcl


proc updateTitle {} {
    global modulesString pollTime

    wm title . "moodss: $modulesString data (every $pollTime seconds)"                 ;# show modules and poll time in window title
}

proc createMenuWidget {parentPath} {
    global modules

    set menu [menu $parentPath.menu -tearoff 0]

    $menu add cascade -label File -menu [menu $menu.file -tearoff 0] -underline 0
    $menu.file add command -label Save -command {save 0} -underline 1 -accelerator Alt+S
    bind $parentPath <Alt-s> {save 0}
    $menu.file add command -label {Save As...} -command {save 1} -underline 6 -accelerator Alt+A
    bind $parentPath <Alt-a> {save 1}
    $menu.file add command -label Exit -command exit -underline 1 -accelerator Alt+X
    bind $parentPath <Alt-x> exit

    $menu add cascade -label Options -menu [menu $menu.options -tearoff 0] -underline 0
    $menu.options add command -label {Poll Time...} -command inquirePollTime -underline 0

    $menu add cascade -label Help -menu [menu $menu.help -tearoff 0] -underline 0
    $menu.help add command -label Global... -underline 0 -accelerator F1 -command helpWindow
    bind $parentPath <F1> helpWindow
    $menu.help add command -label Modules... -underline 0 -command "modulesHelpDialogBox $modules"
    $menu.help add command -label Copyright... -underline 0 -command {simpleTextDialogBox {moodss: Copyright} $help::copyright}
    $menu.help add command -label About... -underline 0 -command aboutDialogBox

    $parentPath configure -menu $menu
}

proc createMessageWidget {parentPath} {
    global messenger

    set messenger [new lifoLabel $parentPath -headerfont $font::(mediumBold) -font $font::(mediumNormal)]
    return $widget::($messenger,path)                                                                     ;# return actual tk widget
}

proc createDragAndDropZone {parentPath} {
    global canvas pollTime

    set frame [frame $parentPath.drops]

    set label [label $frame.graph -image [image create photo -data [dataGraph::iconData]]]
    pack $label -pady 1 -side left
    new dropSite -path $label -formats DATACELLS\
        -command "viewer::view \[new dataGraph $canvas -interval $pollTime\] \$dragSite::data(DATACELLS)"
    new widgetTip -path $label -text {graph drop site}

if {$::officialBLT} {
    set label [label $frame.overlapBarChart -image [image create photo -data [dataOverlapBarChart::iconData]]]
    pack $label -pady 1 -side left
    new dropSite\
        -path $label -formats DATACELLS -command "viewer::view \[new dataOverlapBarChart $canvas\] \$dragSite::data(DATACELLS)"
    new widgetTip -path $label -text {overlap bar chart drop site}
}

    set label [label $frame.sideBarChart -image [image create photo -data [dataSideBarChart::iconData]]]
    pack $label -pady 1 -side left
    new dropSite\
        -path $label -formats DATACELLS -command "viewer::view \[new dataSideBarChart $canvas\] \$dragSite::data(DATACELLS)"
    new widgetTip -path $label -text {side bar chart drop site}

    set label [label $frame.stackedBarChart -image [image create photo -data [dataStackedBarChart::iconData]]]
    pack $label -pady 1 -side left
    new dropSite -path $label -formats DATACELLS\
        -command "viewer::view \[new dataStackedBarChart $canvas\] \$dragSite::data(DATACELLS)"
    new widgetTip -path $label -text {stacked bar chart drop site}

    set label [label $frame.2DPieChart -image [image create photo -data [data2DPieChart::iconData]]]
    pack $label -pady 1 -side left
    new dropSite -path $label -formats DATACELLS\
        -command "viewer::view \[new data2DPieChart $canvas\] \$dragSite::data(DATACELLS)"
    new widgetTip -path $label -text {2D pie chart drop site}

    set label [label $frame.3DPieChart -image [image create photo -data [data3DPieChart::iconData]]]
    pack $label -pady 1 -side left
    new dropSite -path $label -formats DATACELLS\
        -command "viewer::view \[new data3DPieChart $canvas\] \$dragSite::data(DATACELLS)"
    new widgetTip -path $label -text {3D pie chart drop site}

    set trashData {
        R0lGODlhKAAoAKUAAP8A/8/Lz+/r71FRUZ6anpaSlv/7/wgICL66vkFBQYaChtfT17LA3LaytjAw
        MFlZWZaant/b3/fz96aipnFxcWlpaWFhYTAoKCAgICgoKOfj576yvklJSRAQEBAQCI6Slq6yrp6S
        nhgYGAAAAI6KjoaKhnl5ecfDx66ipiAoIOfr5ygwMJ6Slnlxce/j5zg4OL7Dx6aint/T38fLz4Z5
        hvf7/8/Tz66qtqairv/z939/f/fz79fLz0lRUXmChllRUSH5BAEAAAAALAAAAAAoACgAAAb+QIBw
        SBwGisikckkUDAhMYiEaNQwO06ggS00iDIXEQcFcLA5dpUDAaCgch8GDDIVEDJJJIp2URBQUFQoW
        CRcYGRkGihoBCnt8RQYbDxwZhx0eBwcYHyAMIRMDDpBEEwYPGBgiHSOtrh0PJCVhGaRDCw0JHq68
        vAkDJh0cfBIEHAoNGb3LvR0ODlBDJwsTSHcnBRaszNytAxMaa4oGJyRCAn4LBggMBAfd8BwQCxL1
        AiiPAOrjDRMU8PBSUDghIJyKFcOEBDAQIMKJCQqUAeQW54QEDREKjOACQIOGAhQ4iJgY0KIAFgco
        FFlIgoKDXSS7UZAggIKIJAJCtFjxLib+NwcuNLwYkARGDAovevpchkFGhBFLEJhIupTbDA2jlBho
        8K/qsg400CzJ80Cp11Ydhi6JEIBEgm1nW2HgkBDJFwQVHGCI6wqRWGs1ClRIMJLviAQmoCbBY4LD
        i1R8LUSwYSJJAAE3OhzogDBuBgISYNRCouFEhgHmAFCAWbUCugkvcKLCMYREA7MxNRjQMCFrkRMY
        og0xUCGFzwMcDORYMImYBhYJJHbTgUPGDow8eORKI2AHAkeqzB54QYG6ABkCFgSYESDACOFM0GGr
        8MCCBQo9FJgowAAHjjUCGODCAjxE0EAHFnSh2wkfkOADDSGwoIA/CTgwgH8MKKCAAC5S8MBAASLk
        w8QJBsgwAQEk0PBDYX4dgAMCrfg2AUqjUaGCAQsgQMIKRCHxT4K2EEEAOjZYIAJHRKQWJBEgCJCD
        kUtCgoJRHkQJCQkWbGTllrYEAQA7
    }
    set label [label $frame.trash -image [image create photo -data $trashData]]
    pack $label -pady 1 -side right
    new dropSite -path $label -formats OBJECTS -command "eval delete \$dragSite::data(OBJECTS)"
    new widgetTip -path $label -text {objects trash drop site}
    return $frame
}

proc inquirePollTime {} {
    global pollTimes pollTime

    set dialog\
        [new dialogBox . -buttons oc -default o -title {moodss: Poll Time} -die 0 -x [winfo pointerx .] -y [winfo pointery .]]
    set frame [frame $widget::($dialog,path).frame]
    set minimum [lindex $pollTimes 0]
    set message [message $frame.message\
        -width [winfo screenwidth .] -font $font::(mediumNormal) -justify center\
        -text "Please enter new poll time\n(greater than $minimum):"\
    ]
    pack $message
    set entry [new spinEntry $frame -width 4 -list $pollTimes -side right]
    spinEntry::set $entry $pollTime
    pack $widget::($entry,path) -anchor e -side left -expand 1 -padx 2       ;# evenly pack entry and label together near the center
    pack [label $frame.label -text seconds] -anchor w -side right -expand 1 -padx 2
    dialogBox::display $dialog $frame
    widget::configure $dialog -command "
        set time \[spinEntry::get $entry\]
        if {\$time<$minimum} {                                                                    ;# check against minimum poll time
            bell
            $message configure -text {Please enter new poll time\n(must be greater than $minimum):}
        } else {                                                                                                ;# new time is valid
            if {\$time!=\$pollTime} {                                             ;# but check that it actually differs from current
                set pollTime \$time
                viewer::updateInterval \$time
                updateTitle
                refresh                                      ;# update immediately in case poll time was set to a much greater value
            }
            delete $dialog                                                                                     ;# destroy dialog box
        }
    "
    bind $frame <Destroy> "delete $entry"                                            ;# delete objects not managed by the dialog box
}

proc save {ask} {                               ;# save current configuration in user defined file or currently defined storage file
    global saveFile messenger

    if {$ask||([string length $saveFile]==0)} {
        set file [tk_getSaveFile\
            -title {moodss: Save} -initialdir [pwd] -defaultextension .moo -filetypes {{{moodss data} .moo}} -initialfile $saveFile
        ]
        if {[string length $file]==0} return                                                                 ;# user canceled saving
        set saveFile $file
    }
    lifoLabel::flash $messenger "saving in $saveFile..."                ;# flash instead of pushing as that goes too fast to be seen
    set record [new record -file $saveFile]
    record::write $record
    delete $record
}

proc refresh {} {
    global updateEvent modulesString modules messenger pollTime

    catch {after cancel $updateEvent}                                                             ;# eventually cancel current event

    lifoLabel::push $messenger "launching $modulesString data update..."        ;# we do not know when data will actually be updated
    update idletasks
    foreach module $modules {
        ${module}::update                                                                   ;# ask module to update its dynamic data
    }
    lifoLabel::pop $messenger
    set updateEvent [after [expr {1000*$pollTime}] refresh]                                               ;# convert to milliseconds
}

createMenuWidget .
pack [createMessageWidget .] -side bottom -fill x
updateTitle

set scroll [new scroll canvas .]
set canvas $composite::($scroll,scrolled,path)
set width [winfo screenwidth .]
set height [winfo screenheight .]
$canvas configure -background white -width $width -height $height -scrollregion [list 0 0 $width $height]

if {[info exists initializer]} {
    foreach {width height} [record::sizes $initializer] {}                                   ;# used stored geometry for main window
} else {
    set width 400
    set height 300
}
wm geometry . ${width}x$height

pack [createDragAndDropZone .] -fill x
pack $widget::($scroll,path) -fill both -expand 1

set x 0
set y 0
foreach module $modules {
    upvar #0 ${module}::data data
    if {[info exists data(views)]} {                                                                   ;# user overrode default view
        set index 0
        foreach members $data(views) {                                                    ;# create and manage a table for each view
            array set view $members
            set table [new dataTable $canvas -data ${module}::data -view ::view]
            unset view
            if {[info exists initializer]} {
                foreach {x y width height} [record::tableGeometry $initializer $module $index] {}  ;# used stored geometry for table
                composite::configure $dataTable::($table,handles) -setx $x -sety $y -setwidth $width -setheight $height
            } else {
                composite::configure $dataTable::($table,handles) -setx $x -sety $y
                incr x $configuration::(xWindowManagerInitialOffset)                      ;# offset tables to achieve a nicer layout
                incr y $configuration::(yWindowManagerInitialOffset)
            }
            incr index                                                                                  ;# next view for initializer
        }
    } else {                                                                                                ;# use default data view
        set table [new dataTable $canvas -data ${module}::data]
        if {[info exists initializer]} {
            foreach {x y width height} [record::tableGeometry $initializer $module] {}             ;# used stored geometry for table
            composite::configure $dataTable::($table,handles) -setx $x -sety $y -setwidth $width -setheight $height
        } else {
            composite::configure $dataTable::($table,handles) -setx $x -sety $y
            incr x $configuration::(xWindowManagerInitialOffset)                          ;# offset tables to achieve a nicer layout
            incr y $configuration::(yWindowManagerInitialOffset)
        }
    }
}

refresh                                                                                                ;# initialize refresh process

if {[info exists initializer]} {                  ;# stored configuration, now that modules data is initialized, create some viewers
    foreach {class cells x y width height} [record::viewersData $initializer] {
        set viewer [new $class $canvas]
        viewer::view $viewer $cells
        composite::configure $viewer::($viewer,handles) -setx $x -sety $y -setwidth $width -setheight $height
    }
}
